# -*- encoding: utf-8 -*-
"""py3o.template exposes a dirt simple API to render templated OpenOffice
documents into real OpenOffice documents with all your data merged-in.
"""

from py3o.template.main import Template
from py3o.template.main import TextTemplate
from py3o.template.main import TemplateException
